import threading
from typing import *
from ..EventArgs import *


class BaseTcpHandler(threading.Thread):
    def __init__(
        self,
        on_message: Callable,
        on_new_connection: Callable = None,
        on_close_connection: Callable = None,
        # on_message: Callable[[MessageEventArgs], None],
        # on_new_connection: Callable[[NewConnectionEventArgs], None] = None,
        # on_close_connection: Callable[[CloseConnectionEventArgs], None] = None,
        daemon: bool = True
    ):
        self.on_message = on_message
        self.on_new_connection = on_new_connection
        self.on_close_connection = on_close_connection
        return super().__init__(daemon=daemon)

    def OnMessage(self, remote_host: str, connection: socket.socket, data: bytes):
        if self.on_message:
            e = MessageEventArgs(
                host=remote_host,
                connection=connection,
                data=data
            )
            self.on_message(e)

    def OnNewConnection(self, remote_host: str, connection: socket.socket):
        if self.on_new_connection:
            e = NewConnectionEventArgs(remote_host, connection)
            return self.on_new_connection(e)

    def OnCloseConnection(self):
        pass



from typing import Callable, Dict


def dict2sheet(data: Dict, formatter: Callable = None, header: dict = {'key': 'value', '-'*20: '-'*30}, margin: int = 30):
    '''
      convert key-value to sheet
      formatter: Callable[line:str,lines_dict:Dict] -> str : specify how to format a line
            default set is lambda x,dic:f'{left_just(x)}\t{hex(dic[x]) if isinstance(dic[x],int) else dic[x]}'
            if formatter is set , no need to set margin
      
    '''
    result = dict(header)
    result.update(data)

    def left_just(x: str):
        x = str(x)
        return x.ljust(margin, ' ')
    if not formatter:
        def formatter(
            x, dic): return f'{left_just(x)}\t{hex(dic[x]) if isinstance(dic[x],int) else dic[x]}'
    output = [formatter(x, result) for x in result]
    return output




import itertools
from typing import Callable, List
from .typing import cast2bytes
def xor(data: any, xor: any) -> bytes:
    data = cast2bytes(data)
    xor = cast2bytes(xor)
    result = [a ^ b for (a, b) in zip(data, itertools.cycle(xor))]
    return cast2bytes(result)


def find(arr: List, predict: Callable) -> any:
    c_count = predict.__code__.co_argcount
    if c_count == 1:
        def x_predict(i, index): return predict(i)
    elif c_count == 2:
        x_predict = predict
    else:
        raise Exception(f'invalid param count {c_count}')
    for index, i in enumerate(arr):
        if x_predict(i, index):
            return i
    return None


def flat(arr: List, rank: int = -1) -> List:
    if rank == 0:
        return arr
    result = []
    for i in arr:
        if isinstance(i, List):
            result += flat(i, rank-1)
        else:
            result.append(i)
    return result


def distinct(arr: List, predict: Callable = None) -> List:
    '''
    distinct a array
    arr     :   List[T]
    predict :   Callable[T,str]
    '''
    if not predict:
        def predict(x): return x
    dic = {}
    result = []
    for i in arr:
        k = predict(i)
        if k in dic:
            continue
        dic[k] = True
        result.append(i)
    return result



from .full_tree_ex import BinaryTree
from typing import List, Type, overload
from .node import Node

class BinaryTree(BinaryTree):
    def __from_inner_order(otheroder: List, inorder: List, order: int = 0) -> BinaryTree:
        if not otheroder or not inorder:
            return None
        l = len(otheroder)
        # 创建根节点，根节点是前序遍历的第一个数，根节点是后序遍历的最后一个数
        root = otheroder[0] if order == 0 else otheroder[l-1]
        node = Node(root)
        index = inorder.index(root)  # 找到中序遍历根节点所在位置

        l_inorder = inorder[:index]
        r_inorder = inorder[index+1:]
        if order == 0:
            # 对于中序遍历，根节点左边的节点即左子树，根节点右边的节点即右子树
            l_oth = otheroder[1:index+1]
            r_oth = otheroder[index+1:]
            node.lchild = BinaryTree.__from_inner_order(
                l_oth, l_inorder, order)
            node.rchild = BinaryTree.__from_inner_order(
                r_oth, r_inorder, order)
        else:
            l_oth = otheroder[:index]
            r_oth = otheroder[index:l-1]
            node.lchild = BinaryTree.__from_inner_order(
                l_oth, l_inorder, order)
            node.rchild = BinaryTree.__from_inner_order(
                r_oth, r_inorder, order)
        node.data = root
        return node

    def from_inner_order(otheroder: List, inorder: List, order: int = 0) -> BinaryTree:
        '''
        通过两个顺序还原二叉树
        order:int
          0:通过前序和中序获取
          1:通过后序和中序获取
        '''
        t = type(otheroder)
        otheroder = BinaryTree.__convert_data(otheroder)
        inorder = BinaryTree.__convert_data(inorder)
        r = BinaryTree.__from_inner_order(otheroder, inorder, order)
        return BinaryTree(r, as_type=t)

    def from_preorder_and_inorder(preorder: List, inorder: List) -> BinaryTree:
        return BinaryTree.from_inner_order(preorder, inorder, 0)

    def from_postorder_and_inorder(postorder: List, inorder: List) -> BinaryTree:
        return BinaryTree.from_inner_order(postorder, inorder, 1)




import inspect
from .openpyxl_extend import *
import openpyxl
from openpyxl.worksheet.worksheet import Worksheet
# 用于对Excel文件进行操作

import os


class XlsxHandler:
    use_new_sheet = True
    wb: openpyxl.Workbook = None

    def __init__(self, target_file: str):
        self.target_file = target_file
        if os.path.exists(self.target_file):
            self.wb = openpyxl.load_workbook(
                filename=self.target_file)  # 缓存excel文件对象
        else:
            print(f'warn:file not exist:{self.target_file},use in memory')
            self.wb = openpyxl.Workbook()


    def read(self, sheet: str = None, headers: dict = None, mapper: dict = None, value_convert: dict = None):
        '''
        将表格转换成为list类型方便后续操作
        sheet		:string	需要读取的表名称,默认读取第一个表
        headers	:dict	需要读取的表头名称,默认读全部列
        mapper    :dict   将原始表头转换为对应的key
        value_convert :dict 实现excel输入项的值转换
        '''
        sh = self.wb[0] if sheet is None else self.wb[sheet]
        sh_headers = self.__get_headers_from_sheet(sh, headers, mapper)
        return self.__read_data(sh, sh_headers, value_convert)

    def __read_data(self, sh: Worksheet, headers: dict, value_convert: dict):
        if not value_convert:
            value_convert = {}
        any_data = {}
        result = [{} for x in range(sh.max_row)]  # 初始化数据
        for row in range(1, sh.max_row+1):
            for col in headers:
                item = headers[col]
                v = sh.cell(row+1, item['index']).value  # 从第2行开始读取
                name = item['name']
                if v:
                    any_data[row] = True
                    # 将当前行数据入内存
                    result[row-1][name] = XlsxHandler.__convert_data(
                        v, value_convert.get(name))
        # 排除无效行
        return [x for index, x in enumerate(result) if index+1 in any_data]

    def __convert_data(data: str, converter):
        if not converter or data is None:
            return data
        if inspect.isfunction(converter):
            return converter(data)
        if converter == 'list':
            return data.split(',')
        return data

    

    def write(self, data: dict, sheet: str = None, append: bool = False):
        '''
        将数据写入到表格中
        sheet		:string	需要读取的表名称,创建下一个表
        data      :Object 需要存入的数据
        '''
        sh, is_new_sheet = get_sheet_with_create(
            self.wb, sheet, put_index=0, use_new_sheet=not append and self.use_new_sheet)
        headers = self.__get_headers_from_data(data)
        self.__build_headers(sh, headers)
        self.__build_values(sh, data, headers, append)
        self.wb.save(self.target_file)

    def __build_values(self, sh: Worksheet, data: dict, headers: list, append: bool = False):
        current_index = 1 if not append else sh.max_row
        current_index += 1  # 从下一行开始
        for index, item in enumerate(data):
            for h_index, header in enumerate(headers):
                if not header in item:
                    continue
                v = XlsxHandler.__to_value(item[header])
                sh.cell(index+current_index, h_index + 1).value = v

    def __to_value(value) -> str:
        if isinstance(value, int) or isinstance(value, float) or isinstance(value, str):
            return value
        if isinstance(value, list):
            return ','.join([str(x) for x in value])
        return str(value)

    def __build_headers(self, sh, headers):
        for index, h in enumerate(headers):
            sh.cell(1, index+1).value = h

    def __get_headers_from_data(self, data):
        result = {}
        for item in data:
            for k in item:
                result[k] = True
        return list(result)

    def __get_headers_from_sheet(self, sh, headers=None, mapper=None):
        if mapper is None:
            mapper = {}
        sh_headers = {}  # 加载header的列索引
        for col in range(1, sh.max_column+1):
            h = sh.cell(1, col).value
            if h is None:
                continue
            if not headers is None and not h in headers:
                continue
            if h in sh_headers:
                print(f'warn:sheet_{sh.title}_{h} is duplicated')
            if h in mapper:
                h = mapper[h]
            sh_headers[h] = {
                'index': col,
                'name': h
            }
        return sh_headers




from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5 as PKCS1_signature
from Crypto.Cipher import PKCS1_v1_5 as PKCS1_cipher
from Crypto.Hash import SHA


def main(content: bytes):
    rsa = RSA.generate(2048)
    public_key = rsa.public_key().export_key()
    private_key = rsa.export_key()
    print(f'public_key: {public_key}\n\nprivate_key: {private_key}\n')
    private_key = RSA.importKey(private_key)
    public_key = RSA.importKey(public_key)

    # sign a message
    signer = PKCS1_signature.new(private_key)
    digest = SHA.new()
    digest.update(content)
    sign = signer.sign(digest)
    print(f'sign: {sign}\n')

    # check sign for a message
    verifier = PKCS1_signature.new(public_key)
    digest = SHA.new()
    digest.update(content)
    sign_result = verifier.verify(digest, sign)
    print(f'the message check sign result is :{sign_result}\n')

    # encrypt rsa
    cipher = PKCS1_cipher.new(public_key)
    encrypt_text = cipher.encrypt(content)
    print(f'encrypt_text: {encrypt_text}\n')

    # decrypt rsa
    cipher = PKCS1_cipher.new(private_key)
    decrypt_text = cipher.decrypt(encrypt_text, b'')
    print(f'decrypt_text: {decrypt_text}\n')



import json
from ..SimpleTcpService import SimpleTcpService, MessageEventArgs
from ...extensions import clazz


class SimpleCaculateService(SimpleTcpService):
    def handle_caculate(self, event: MessageEventArgs):
        data = event.data
        results = {}
        result = {}
        try:
            data = data.decode()
            first_line = data.find('\n')
            # have at least 2 lines
            if first_line > -1 and data.find('\n', first_line+1) > -1:
                exec(data, results)
                keys = [x for x in results if not x.startswith('_')]
                for x in keys:
                    result[x] = results[x]
                result = json.dumps(result)
            else:
                result = eval(data)
        except Exception as e:
            result = f'Error: {e}'
        event.connection.send(f'{result}\n'.encode())

    def __init__(self, host: str = '127.0.0.1'):
        return super().__init__(
            host=host,
            on_message=self.handle_caculate
        )



from enum import Enum
import hashlib

ALG_CLASS_ANY = 0
ALG_CLASS_SIGNATURE = 8192
ALG_CLASS_MSG_ENCRYPT = 16384
ALG_CLASS_DATA_ENCRYPT = 24576
ALG_CLASS_HASH = 32768
ALG_CLASS_KEY_EXCHANGE = 40960
ALG_CLASS_ALL = 57344  # / * (7 << 13) * /
ALG_TYPE_ANY = 0
ALG_TYPE_DSS = 512
ALG_TYPE_RSA = 1024
ALG_TYPE_BLOCK = 1536
ALG_TYPE_STREAM = 2048
ALG_TYPE_DH = 2560  # / * (5 << 9) * /
ALG_TYPE_SECURECHANNEL = 3072  # / * (6 << 9) * /
ALG_SID_ANY = 0
ALG_SID_RSA_ANY = 0
ALG_SID_RSA_PKCS = 1
ALG_SID_RSA_MSATWORK = 2
ALG_SID_RSA_ENTRUST = 3
ALG_SID_RSA_PGP = 4
ALG_SID_DSS_ANY = 0
ALG_SID_DSS_PKCS = 1
ALG_SID_DSS_DMS = 2
ALG_SID_DES = 1
ALG_SID_3DES = 3
ALG_SID_DESX = 4
ALG_SID_IDEA = 5
ALG_SID_CAST = 6
ALG_SID_SAFERSK64 = 7
ALG_SID_SAFERSK128 = 8
ALG_SID_3DES_112 = 9
ALG_SID_SKIPJACK = 10
ALG_SID_TEK = 11
ALG_SID_CYLINK_MEK = 12
ALG_SID_RC5 = 13
ALG_SID_RC2 = 2
ALG_SID_RC4 = 1
ALG_SID_SEAL = 2
ALG_SID_MD2 = 1
ALG_SID_MD4 = 2
ALG_SID_MD5 = 3
ALG_SID_SHA = 4
ALG_SID_MAC = 5
ALG_SID_RIPEMD = 6
ALG_SID_RIPEMD160 = 7
ALG_SID_SSL3SHAMD5 = 8
ALG_SID_HMAC = 9
ALG_SID_TLS1PRF = 10
ALG_SID_AES_128 = 14
ALG_SID_AES_192 = 15
ALG_SID_AES_256 = 16
ALG_SID_AES = 17
ALG_SID_EXAMPLE = 80

CALG_MD2 = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_MD2)
CALG_MD4 = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_MD4)
CALG_MD5 = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_MD5)
CALG_SHA = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_SHA)
CALG_SHA1 = CALG_SHA
CALG_MAC = (ALG_CLASS_HASH | ALG_TYPE_ANY | ALG_SID_MAC)
CALG_3DES = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_BLOCK | 3)
CALG_CYLINK_MEK = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_BLOCK | 12)
CALG_SKIPJACK = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_BLOCK | 10)
CALG_KEA_KEYX = (ALG_CLASS_KEY_EXCHANGE | ALG_TYPE_STREAM | ALG_TYPE_DSS | 4)
CALG_RSA_SIGN = (ALG_CLASS_SIGNATURE | ALG_TYPE_RSA | ALG_SID_RSA_ANY)
CALG_DSS_SIGN = (ALG_CLASS_SIGNATURE | ALG_TYPE_DSS | ALG_SID_DSS_ANY)
CALG_RSA_KEYX = (ALG_CLASS_KEY_EXCHANGE | ALG_TYPE_RSA | ALG_SID_RSA_ANY)
CALG_DES = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_BLOCK | ALG_SID_DES)
CALG_RC2 = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_BLOCK | ALG_SID_RC2)
CALG_RC4 = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_STREAM | ALG_SID_RC4)
CALG_SEAL = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_STREAM | ALG_SID_SEAL)
CALG_DH_EPHEM = (ALG_CLASS_KEY_EXCHANGE | ALG_TYPE_STREAM |
                 ALG_TYPE_DSS | ALG_SID_DSS_DMS)
CALG_DESX = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_BLOCK | ALG_SID_DESX)
CALG_AES_128 = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_BLOCK | ALG_SID_AES_128)
CALG_AES_192 = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_BLOCK | ALG_SID_AES_192)
CALG_AES_256 = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_BLOCK | ALG_SID_AES_256)
CALG_AES = (ALG_CLASS_DATA_ENCRYPT | ALG_TYPE_BLOCK | ALG_SID_AES)

print('CALG_MD5', hex(CALG_MD5))
print('CALG_SHA1', hex(CALG_SHA1))


class HashAlgo(Enum):
    md5 = CALG_MD5
    sha1 = CALG_SHA1
    sha224 = 'sha224'
    sha256 = 'sha256'
    sha384 = 'sha384'
    sha512 = 'sha512'


def get_hash(content: bytes, hash_name: HashAlgo = HashAlgo.md5, key: bytes = None) -> str:
    '''
      get hexdigest of a content

      Arguments:
        content:bytes: str|bytes
        hash_name: HashAlgo: named of hash
        key:bytes: if use rc4 etc. , key should be added
    '''
    if isinstance(hash_name, HashAlgo):
        hash_name = hash_name.name
    assert hasattr(hashlib, hash_name), 'no such hash algorithm'
    f = getattr(hashlib, hash_name)
    model = f()
    if isinstance(content, str):
        content = content.encode('ascii')
    model.update(content)
    return model.hexdigest()



import openpyxl


def exist(wb: openpyxl.Workbook, sheet_name: str) -> bool:
    return sheet_name in _sheets_dict(wb)


def _sheets_dict(wb: openpyxl.Workbook) -> dict:
    wb._sheets_dict = {}
    for x in wb._sheets:
        wb._sheets_dict[x.title] = x
    return wb._sheets_dict


def get_sheet_with_create(wb: openpyxl.Workbook, title: str = None, index: int = None, put_index: int = None, use_new_sheet: bool = False, prefix_index: int = None):
    """get a worksheet (at an optional index).
    and create one if not exist

    :param title: optional title of the sheet
    :type title: str
    :param index: optional position at which the sheet will be inserted
    :type index: int

    :return sheet,is_new_sheet
    """
    if not title is None:
        current_title = title if not prefix_index else f'{prefix_index}_{title}'
        e = exist(wb, current_title)
        if e:
            if use_new_sheet:
                return get_sheet_with_create(wb, title, index, put_index, use_new_sheet, prefix_index=prefix_index+1 if prefix_index else 1)
            return (wb[title], False)
        return (wb.create_sheet(title, put_index), True)
    if not index is None:
        if index >= 0 and index < len(wb.sheets):
            (wb.sheets[index], False)
    s_name = f'new_sheet_{len(wb.sheets)}'
    return (wb.create_sheet(s_name, put_index), True)




from ..logger import logger
import json
import os
from typing import Dict
configuration_path: str = '.sgt.conf'
config: Dict = None


def get_config_path() -> str:
    global configuration_path
    path = os.path.join(os.path.expanduser('~'), configuration_path)
    return path


def load(config_path: str = None, reload: bool = False) -> Dict:
    if not config_path is None:
        global configuration_path
        if config_path != configuration_path:
            reload = True
        configuration_path = config_path
    global config
    if config is None:
        reload = True
    if not reload:
        return config
    p = get_config_path()
    if not os.path.exists(p):
        save(reload=False)
    with open(p, 'r', encoding='utf-8') as f:
        config_content = f.read()
    config = json.loads(config_content) or {}
    return config


def save(new_config: Dict = None, reload: bool = False):
    global config
    data = json.dumps(config if new_config is None else new_config)
    with open(get_config_path(), 'w', encoding='utf-8') as f:
        f.write(data)
    if reload:
        load()


def get(key: str = None, default: any = None) -> any:
    data = load()
    if not key in data:
        return default
    return data[key]


def set(key: str, value: any = None):
    load()
    global config
    key = str(key)
    if not key:
        raise Exception('key must be set')
    config[key] = value



import logging
import logging.handlers
import colorlog
logger: logging.Logger = None
file_handler: logging.Handler = None
console_handler: logging.Handler = None

LOG_FILE = 'log.log'


def set_log_file(filename: str):
    global LOG_FILE
    LOG_FILE = filename
    init()


def disable(file: bool = True, console: bool = True):
    global logger
    global file_handler
    global console_handler
    if not logger:
        return
    if file:
        logger.removeHandler(file_handler)
    if console:
        logger.removeHandler(console_handler)


def redefine_level_name():
    from logging import CRITICAL, FATAL, ERROR, WARNING, WARN, INFO, DEBUG, NOTSET
    dic = {
        CRITICAL: '!!',
        FATAL: '!!',
        ERROR: '-',
        WARNING: '!',
        WARN: '!',
        INFO: '+',
        DEBUG: '*',
        NOTSET: '~',
    }
    for i in dic:
        logging.addLevelName(i, dic[i])


redefine_level_name()


def init():
    global LOG_FILE
    global logger
    global file_handler
    global console_handler
    logger = logging.getLogger('common')  # 获取名为commomn的logger。
    fmt = '%(asctime)s [%(levelname)s] %(message)s :%(filename)s:%(lineno)s-%(levelno)s  %(pathname)s on(%(module)s.%(funcName)s) at %(process)d@%(threadName)s  '  # 定义日志格式
    file_handler = logging.handlers.RotatingFileHandler(
        LOG_FILE, maxBytes=1024*1024, backupCount=5)
    formatter = logging.Formatter(fmt)   # 实例化formatter。
    file_handler.setFormatter(formatter)      # 为handler添加formatter。
    logger.addHandler(file_handler)           # 为logger添加handler。

    fmt = '%(asctime)s [%(levelname)s] %(message)s'  # 定义日志格式
    colors = {
        '*': 'light_blue',
        '+': 'green',
        '!': 'yellow',
        '-': 'red',
        '!!': 'purple,bg_white',
    }
    fmt_colored = colorlog.ColoredFormatter(
        f'%(log_color)s{fmt}', datefmt=None, reset=True, log_colors=colors)
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(fmt_colored)      # 为handler添加formatter。
    logger.addHandler(console_handler)           # 为logger添加handler。

    logger.setLevel(logging.DEBUG)


init()



import inspect
from typing import Callable, Dict, List


def convert_to_value(i: object, clazz: type, callback: Callable):
    '''
    将i转换为目标值
    '''
    if isinstance(i, int):
        return i
    elif clazz == i.__class__:
        return callback(i)
    else:
        raise Exception('invalid input')


def get_data(item: object, fields_mapper: List) -> Dict:
    '''
    get data from a object with fields specified
    '''
    result = {}
    tstring = type('')
    tarray = type([])
    for i in fields_mapper:
        ti = type(i)
        if ti == tstring:
            result[i] = getattr(item, i)
        elif ti == tarray:
            result[i[0]] = i[1](item)
        else:
            raise Exception('invalid operation')
    return result


def class2props(obj: object):
    '''
    将class转dict,以_开头的属性不要
    '''
    if isinstance(obj, dict):
        return obj
    pr = {}
    for name in dir(obj):
        value = getattr(obj, name)

        if not name.startswith('__') and not inspect.isbuiltin(value) and not inspect.isfunction(value) and not name.startswith('_'):
            pr[name] = value
    return pr


def class2props_with_(obj: object) -> Dict:
    '''
    将class转dict,以_开头的也要
    '''
    if isinstance(obj, dict):
        return obj
    pr = {}
    for name in dir(obj):
        value = getattr(obj, name)
        if not name.startswith('__') and not inspect.isbuiltin(value) and not inspect.isfunction(value):
            pr[name] = value
    return pr


def dict2obj(obj: object, dict: Dict) -> object:
    '''
    dict转obj，先初始化一个obj
    '''
    obj.__dict__.update(dict)
    return obj



import inspect
import sys


class T:
    TA: str = None


def setups(target: type, names: list):
    [setup(target, x) for x in names]


def setup(target: type, name: str):
    n = '__name__'

    def PredicateProperty(v):
        r = type(v) == type(T.TA)
        if not r:
            return r
        if not hasattr(v, n) or getattr(v, n) in ['__doc__']:
            return False
        return r

    def PredicateFunction(v):
        r = type(v) == type(setup)
        if not r:
            return False
        if getattr(v, n) in ['setup', 'setups']:
            return False
        return True

    def MemberPredict(v):
        r = PredicateProperty(v)
        r |= PredicateFunction(v)
        return r
    if not name in sys.modules:
        __import__(name)
    m = sys.modules[name]
    members = inspect.getmembers(m, MemberPredict)
    [setattr(target, x[0], x[1]) for x in members]



from __future__ import annotations
import string
from typing import Dict, List, Tuple, overload
from ..logger import logger


class Pos(Tuple):
    @overload
    def __new__(self, pos: Pos) -> Pos:
        ...

    @overload
    def __new__(self, x: int = 0, y: int = 0) -> Pos:
        ...

    def __new__(self, x: any = None, y: int = None) -> Pos:
        if x is None:
            x = 0
            y = 0
        if isinstance(x, Pos):
            x = x.x
            y = x.y
        return super().__new__(self, [x, y])

    @property
    def x(self):
        return self[0]

    @property
    def y(self):
        return self[1]

    @x.setter
    def x(self, val):
        self[0] = val

    @y.setter
    def y(self, val: int):
        self[1] = val

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, Pos):
            return self.x == __o.x and self.y == __o.y
        return False

    def equal(self, x: int, y: int) -> bool:
        # speed up
        return self.x == x and self.y == y

    def move(self, x: int = 0, y: int = 0):
        self.x += x
        self.y += y

    def __repr__(self) -> str:
        return f'Pos[{self.x},{self.y}]'


class Maze:
    default_directions = {
        'a': [-1, 0],
        'd': [1, 0],
        'w': [0, -1],
        's': [0, 1],
    }

    def get_index_by_pos(self, pos: Pos) -> int:
        return pos.x + pos.y * self.map_sizex

    def get_pos_by_index(self, index: int) -> Pos:
        x = index % self.map_sizex
        y = index // self.map_sizex
        return Pos(x, y)

    def __init__(self, map_data: List, map_sizex: int, start: Pos | any, end: Pos | any, avoid: List[any] = None, allow: List[any] = None, directions: Dict = default_directions, direction_char_convert: Dict[any, any] = None, max_result_count: int = 1):
        '''
        map_data:       List draw a map
        map_sizex:      int pos-x of the map , to determine map's shape
        start:Pos:      player start position or start-pos-data , if set data , then will find in map_data and convert to pos
        end:Pos         player end position or end-pos-data , if set pos , then will convert pos to a data and modify map_data
        avoid:List[any] when moving position's data in this , player will die. `avoid` and `allow` should only set one of them
        allow:List[any] when moving position's data in this , nothing happend. `avoid` and `allow` should only set one of them
        directions:Dict determine which direction is allowed to move , default is `up` `down` `left` `right`
        max_result_count:int if get enough result , then directly ret
        direction_char_convert:Dict:you can use converter for convert default `w` to `up` etc.
        '''
        is_data_str = False
        if isinstance(map_data, int):
            map_data = str(map_data)
        if isinstance(map_data, str):
            self.data = list(map_data)
            is_data_str = True
        elif isinstance(map_data, List):
            self.data = map_data
        else:
            raise Exception('map_data should be str')
        self.map_sizex = map_sizex
        self.map_sizey = len(self.data) // self.map_sizex
        map_data_dict = {}
        for index, _ in enumerate(self.data):
            key = self.data[index]
            if key in map_data_dict:
                continue
            map_data_dict[key] = index
        if isinstance(start, Pos):
            self.start = start
        else:
            assert start in map_data_dict, f'start data [{start}] not found in map_data'
            index = map_data_dict[start]
            self.start = self.get_pos_by_index(index)
            logger.info(f'self start-pos\'s pos is set to {self.start}')

        if isinstance(end, Pos):
            for i in string.printable:
                if not i in map_data:
                    self.end = i
                    index = self.get_index_by_pos(end)
                    self.data[index] = i
                    logger.info(f'end-pos\'s data is set to char {i} at {end}')
                    break
        else:
            self.end = end
        self.directions = directions or Maze.default_directions
        self.direction_char_convert = direction_char_convert or {}
        assert not (
            avoid and allow), '`avoid` and `allow` should set one of them'
        assert avoid or allow, '`avoid` and `allow` should set one of them'

        if allow:
            for i in allow:
                if i in map_data_dict:
                    del map_data_dict
            avoid = map_data_dict

        avoid = list(avoid)
        avoid_dict = {}
        for i in avoid:
            avoid_dict[i] = True
        self.avoid = avoid_dict
        self.max_result_count = max_result_count
        if is_data_str:
            self.data = ''.join(self.data)
        self.init_solution()
        pass

    def init_solution(self):
        self.visited: Dict[str, bool] = {}
        self.paths: List[str] = []
        self.result_count = 0
        self.dump_result = None

    def dump_map(self, convert: Dict[any, any] = {}):
        result = []
        r = []
        for index, i in enumerate(self.data):
            if not (index % self.map_sizex):
                result.append(r)
                r = []
            data = self.data[index:index+1]
            if data in convert:
                data = convert[data]
            r.append(data)
        result.append(r)
        return result

    def print_map(self, convert: Dict[any, any] = {}):
        data = [''.join([''.join([str(c) for c in y]) for y in x])
                for x in self.dump_map(convert)]
        print('\n'.join(data))

    def dump(self):
        if self.dump_result:
            return self.dump_result
        if not self.result:
            self.result = self.explore()
        assert self.result, 'fail to solve'
        self.dump_result = []
        self.__dump(self.result, [])
        return self.dump_result

    def __dump(self, dic: Dict, r: List):
        if dic == None or dic == True:
            self.dump_result.append(''.join([str(x) for x in r]))
            return
        for k in dic:
            r.append(k)
            self.__dump(dic[k], r)
            r.pop()

    def explore(self):
        self.init_solution()
        x, y = self.start
        self.result = self.__explore(x, y)
        return self.result

    def __explore(self, x: int, y: int):
        k = f'{x}:{y}'
        if x < 0 or y < 0 or x >= self.map_sizex or y >= self.map_sizey:
            return None
        if k in self.visited:
            return None
        p = self.map_sizex*y+x
        v = self.data[p]
        if v == self.end:
            self.result_count += 1
            return True
        if v in self.avoid:
            return None
        result = {}
        self.visited[k] = True
        for d in self.directions:
            # paths.append('a')
            dx, dy = self.directions[d]
            r = self.__explore(x+dx, y+dy)
            if r:
                if d in self.direction_char_convert:
                    d = self.direction_char_convert[d]
                result[d] = r
            # paths.pop()
            if self.result_count >= self.max_result_count:
                break

        del self.visited[k]
        return result



import socket


class ConnectionEventArgs:
    def __init__(self, host: str, connection: socket.socket):
        self.host = host
        self.connection = connection


class MessageEventArgs(ConnectionEventArgs):
    '''
    a new message distribution
    '''

    def __init__(self, host: str, connection: socket.socket, data: bytes):
        self.data = data
        super().__init__(
            host=host,
            connection=connection
        )


class NewConnectionEventArgs(ConnectionEventArgs):
    pass


class CloseConnectionEventArgs(ConnectionEventArgs):
    pass



def padding_to(raw: bytes, padding: int, max_length: int = None):
    '''
    sometimes crypto need padding length to its requires
    '''
    c = len(raw)/padding
    block: int = int(c)
    if block != c:
        block += 1
    result = raw.rjust(padding*block, b'\0')
    if max_length:
        result = result[0:max_length]
    return result


__title__ = "sgtpyutils"
__description__ = "sgtpyutils is a public library for use."
__keywords__ = ['lib', 'utils']
__url__ = "https://github.com/serfend/sgtpyutils"
__version__ = "1.15.10"
__author__ = "serfend"
__author_email__ = "serfend@foxmail.com"
__license__ = "MIT Licence"

__public_path__ = './dist'



import threading
from typing import *
import socket
from ..EventArgs import *
from ....logger import logger
from ..BaseTcpHandler import BaseTcpHandler

class MessageHanlder(threading.Thread):
    def __init__(self,
                 connection: socket.socket,
                 remote_host: str,
                 callback: BaseTcpHandler,
                 daemon: bool = False) -> None:
        self.running = False

        self.connection = connection
        self.remote_host = remote_host
        self.callback = callback
        return super().__init__(daemon=daemon)

    def run(self) -> None:
        self.running = True
        super().run()
        self.__handle_message()
        self.running = False

    def __handle_message(self) -> None:
        connection = self.connection
        remote_host = self.remote_host
        while self.running:
            data = connection.recv(1024)
            if not data:
                break
            self.callback.OnMessage(
                connection=connection,
                remote_host=remote_host,
                data=data
            )
        try:
            connection.close()
        except:
            pass
        self.callback.OnCloseConnection()
        try:
            logger.debug(f'TCPService:: connection closed from {remote_host}')
        except:
            pass



from .node import Node
from typing import List, Type, overload
from ..extensions import flat

class BinaryTree():

    def __init__(self, root=None, as_type: Type = None):
        self.root = root
        self.as_type = as_type or (type(root) if root else List)
        pass

    @staticmethod
    def __convert_data(data) -> List:
        '''
            conver data to List
        '''
        if not data:
            return None
        if isinstance(data, str) or isinstance(data, bytes):
            data = [x for x in data]
        return list(data)

    def __convert_list(self, data_list: List) -> any:
        '''
            convert list to data determine by `as_type`
        '''
        if self.as_type == List:
            return data_list
        if self.as_type == str:
            return ''.join([str(x) for x in data_list])
        if self.as_type == bytes:
            return b''.join(bytes(x) for x in data_list)
        return data_list

    def __travel(self, node: Node, order: int = 0) -> List[Node]:
        if node is None:
            return None
        left = self.__travel(node.lchild, order)
        right = self.__travel(node.rchild, order)
        r = [left, right]
        r.insert(order, node)
        result = [x for x in flat(r) if x]
        return result  # concat result

    def travel(self, node: Node = None, order: int = 0) -> List:
        if node is None:
            node = self.root
        result = self.__travel(node, order)
        result = [x.data for x in result]
        return self.__convert_list(result)

    def travel_preorder(self, node: Node = None) -> List:
        return self.travel(node, 0)

    def travel_inorder(self, node: Node = None) -> List:
        return self.travel(node, 1)

    def travel_postorder(self, node: Node = None) -> List:
        return self.travel(node, 2)




"""Constants and membership tests for ASCII characters"""

whitespace = ' \t\n\r\v\f'
ascii_lowercase = 'abcdefghijklmnopqrstuvwxyz'
ascii_uppercase = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
ascii_letters = ascii_lowercase + ascii_uppercase
digits = '0123456789'
hexdigits = digits + 'abcdef'
octdigits = '01234567'
punctuation = r"""!"#$%&'()*+,-./:;<=>?@[\]^_`{|}~"""
printable = digits + ascii_letters + punctuation + whitespace


NUL = 0x00  # ^@
SOH = 0x01  # ^A
STX = 0x02  # ^B
ETX = 0x03  # ^C
EOT = 0x04  # ^D
ENQ = 0x05  # ^E
ACK = 0x06  # ^F
BEL = 0x07  # ^G
BS = 0x08  # ^H
TAB = 0x09  # ^I
HT = 0x09  # ^I
LF = 0x0a  # ^J
NL = 0x0a  # ^J
VT = 0x0b  # ^K
FF = 0x0c  # ^L
CR = 0x0d  # ^M
SO = 0x0e  # ^N
SI = 0x0f  # ^O
DLE = 0x10  # ^P
DC1 = 0x11  # ^Q
DC2 = 0x12  # ^R
DC3 = 0x13  # ^S
DC4 = 0x14  # ^T
NAK = 0x15  # ^U
SYN = 0x16  # ^V
ETB = 0x17  # ^W
CAN = 0x18  # ^X
EM = 0x19  # ^Y
SUB = 0x1a  # ^Z
ESC = 0x1b  # ^[
FS = 0x1c  # ^\
GS = 0x1d  # ^]
RS = 0x1e  # ^^
US = 0x1f  # ^_
SP = 0x20  # space
DEL = 0x7f  # delete

controlnames = [
    "NUL", "SOH", "STX", "ETX", "EOT", "ENQ", "ACK", "BEL",
    "BS",  "HT",  "LF",  "VT",  "FF",  "CR",  "SO",  "SI",
    "DLE", "DC1", "DC2", "DC3", "DC4", "NAK", "SYN", "ETB",
    "CAN", "EM",  "SUB", "ESC", "FS",  "GS",  "RS",  "US",
    "SP"
]


def _ctoi(c: str):
    if isinstance(c, str):
        return ord(c)
    else:
        return c


def is_alnum(c: str):
    '''
      a/A - z/Z and 0-9
    '''
    return is_alpha(c) or is_digit(c)


def is_ascii(c: str):
    '''
      below 127
    '''
    return 0 <= _ctoi(c) <= 127


def is_blank(c: str):
    '''
      space and \t
    '''
    return _ctoi(c) in (9, 32)


def is_cntrl(c: str):
    '''
      0-31 and 127
    '''
    return 0 <= _ctoi(c) <= 31 or _ctoi(c) == 127


def is_digit(c: str):
    '''
      0 - 9
    '''
    return 48 <= _ctoi(c) <= 57


def is_graph(c: str):
    '''
      can see
    '''
    return 33 <= _ctoi(c) <= 126


def is_upper(c: str):
    '''
      A - Z
    '''
    return 65 <= _ctoi(c) <= 90


def is_lower(c: str):
    '''
      a - z
    '''
    return 97 <= _ctoi(c) <= 122


def is_alpha(c: str):
    '''
      a - z and A - Z
    '''
    return is_upper(c) or is_lower(c)


def is_print(c: str):
    '''
      can print
    '''
    return 32 <= _ctoi(c) <= 126


def is_punct(c: str):
    '''
      can see and not (a/A - z/Z and 0-9)
    '''
    return is_graph(c) and not is_alnum(c)


def is_space(c: str):
    return _ctoi(c) in (9, 10, 11, 12, 13, 32)


def is_xdigit(c: str):
    return is_digit(c) or (65 <= _ctoi(c) <= 70) or (97 <= _ctoi(c) <= 102)


def is_ctrl(c: str):
    return 0 <= _ctoi(c) < 32


def is_meta(c: str):
    return _ctoi(c) > 127


def ascii(c: str):
    if isinstance(c, str):
        return chr(_ctoi(c) & 0x7f)
    else:
        return _ctoi(c) & 0x7f


def ctrl(c: str):
    if isinstance(c, str):
        return chr(_ctoi(c) & 0x1f)
    else:
        return _ctoi(c) & 0x1f


def alt(c: str):
    if isinstance(c, str):
        return chr(_ctoi(c) | 0x80)
    else:
        return _ctoi(c) | 0x80


def unctrl(c: str):
    bits = _ctoi(c)
    if bits == 0x7f:
        rep = "^?"
    elif is_print(bits & 0x7f):
        rep = chr(bits & 0x7f)
    else:
        rep = "^" + chr(((bits & 0x7f) | 0x20) + 0x20)
    if bits & 0x80:
        return "!" + rep
    return rep



from typing import List, Type, overload
from .node import Node
from .base_tree import BinaryTree as BaseTree
from .full_tree import BinaryTree as FullTree
from .full_tree_ex import BinaryTree as FullTreeEx
from .ordered_tree import BinaryTree as OrderedTree
BinaryTree = OrderedTree



import base64
from . import base58
from . import base91
basexx = {}
basexx['b16decode'] = base64.b16decode
basexx['b32decode'] = base64.b32decode
basexx['b64decode'] = base64.b64decode
basexx['b85decode'] = base64.b85decode
basexx['b32decode'] = base64.b32decode
basexx['b58decode'] = base58.b58decode
basexx['b91decode'] = base91.b91decode

basexx['b16encode'] = base64.b16encode
basexx['b32encode'] = base64.b32encode
basexx['b64encode'] = base64.b64encode
basexx['b85encode'] = base64.b85encode
basexx['b32encode'] = base64.b32encode
basexx['b58encode'] = base58.b58encode
basexx['b91encode'] = base91.b91encode


def switch_base_table_encode(algo_name: str, default_table: str, problem_table: str, content: any):
    if isinstance(content, str):
        print('encode excepted content to be a bytes , assume utf-8 encode.')
        content = content.encode('utf-8')
    if isinstance(problem_table, bytes):
        print('encode excepted problem_table to be a str , assume ascii encode.')
        problem_table = problem_table.decode('ascii')

    if problem_table is None:
        problem_table = default_table
    if len(problem_table) < len(default_table):
        problem_table = str.ljust(
            problem_table, len(default_table), '?')
        print(f'length not enough,padding:\n{problem_table}')

    char_map = str.maketrans(default_table, problem_table)
    if not algo_name in basexx:
        raise Exception(f'no such algorithm {algo_name}')
    f = basexx[algo_name]
    encode_content = f(content)
    if isinstance(encode_content, bytes):
        encode_content = encode_content.decode('ascii')
    result = encode_content.translate(char_map)
    return result


def switch_base_table_decode(algo_name: str, default_table: str, problem_table: str, content: any):
    if isinstance(content, bytes):
        print('decode excepted content to be a str , assume ascii encode.')
        content = content.decode('ascii')
    if isinstance(problem_table, bytes):
        print('decode excepted problem_table to be a str , assume ascii encode.')
        problem_table = problem_table.decode('ascii')
    if problem_table is None:
        problem_table = default_table
    if len(problem_table) < len(default_table):
        problem_table = str.ljust(
            problem_table, len(default_table), '?')
        print(f'length not enough,padding:\n{problem_table}')
    char_map = str.maketrans(problem_table, default_table)
    trans = content.translate(char_map)
    if not algo_name in basexx:
        raise Exception(f'no such algorithm {algo_name}')
    f = basexx[algo_name]
    result = f(trans)
    return result



from Crypto.Cipher import Blowfish
from .utils import padding_to


def test_Blowfish(key, iv, content):
    # MODE_CBC
    e = Blowfish.new(padding_to(key, 32, 32), Blowfish.MODE_CBC,
                     padding_to(iv, 8, 8)).encrypt(padding_to(content, 8))
    d = Blowfish.new(padding_to(key, 32, 32), Blowfish.MODE_CBC,
                     padding_to(iv, 8, 8)).decrypt(e)
    print(f'e = {e}\nd = {d}\n\n')
    # MODE_CFB
    e = Blowfish.new(padding_to(key, 32, 32), Blowfish.MODE_CFB,
                     padding_to(iv, 8, 8)).encrypt(padding_to(content, 8))
    d = Blowfish.new(padding_to(key, 32, 32), Blowfish.MODE_CFB,
                     padding_to(iv, 8, 8)).decrypt(e)
    print(f'e = {e}\nd = {d}\n\n')
    # MODE_ECB no need for iv
    e = Blowfish.new(padding_to(key, 32, 32), Blowfish.MODE_ECB).encrypt(
        padding_to(content, 8))
    d = Blowfish.new(padding_to(key, 32, 32), Blowfish.MODE_ECB).decrypt(e)
    print(f'e = {e}\nd = {d}\n\n')
    # MODE_OFB
    e = Blowfish.new(padding_to(key, 32, 32), Blowfish.MODE_OFB,
                     padding_to(iv, 8, 8)).encrypt(padding_to(content, 8))
    d = Blowfish.new(padding_to(key, 32, 32), Blowfish.MODE_OFB,
                     padding_to(iv, 8, 8)).decrypt(e)
    print(f'e = {e}\nd = {d}\n\n')


def main():
    config = {
        "key": "SGTSerfend2022",
        "iv": "1Ssecret1Ssecret1Ssecret"
    }
    content = ('0123456789abcdefAAAAAAAAAA'*5).encode()
    iv = str(config['iv']).encode()
    key = str(config['key']).encode()
    test_Blowfish(key, iv, content)



from .transform_table import switch_base_table_encode, switch_base_table_decode

base91_default_table = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789!#$%&()*+,./:;<=>?@[]^_`{|}~"'
base85_default_table = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'
base64_default_table = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'
base58_default_table = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
base36_default_table = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
base32_default_table = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'
base16_default_table = '123456789ABCDEF'

def base91_decode(content: str, problem_table: str = None):
    return switch_base_table_decode('b91decode', base91_default_table, problem_table, content)


def base85_decode(content: str, problem_table: str = None):
    return switch_base_table_decode('b85decode', base85_default_table, problem_table, content)


def base64_decode(content: str, problem_table: str = None):
    return switch_base_table_decode('b64decode', base64_default_table, problem_table, content)


def base58_decode(content: str, problem_table: str = None):
    return switch_base_table_decode('b58decode', base58_default_table, problem_table, content)


def base36_decode(content: str, problem_table: str = None):
    return switch_base_table_decode('b36decode', base36_default_table, problem_table, content)


def base32_decode(content: str, problem_table: str = None):
    return switch_base_table_decode('b32decode', base32_default_table, problem_table, content)


def base16_decode(content: str, problem_table: str = None):
    return switch_base_table_decode('b16decode', base16_default_table, problem_table, content)


def base91_encode(content: str, problem_table: str = None):
    return switch_base_table_encode('b91encode', base91_default_table, problem_table, content)

def base85_encode(content: str, problem_table: str = None):
    return switch_base_table_encode('b85encode', base85_default_table, problem_table, content)


def base64_encode(content: str, problem_table: str = None):
    return switch_base_table_encode('b64encode', base64_default_table, problem_table, content)


def base58_encode(content: str, problem_table: str = None):
    return switch_base_table_encode('b58encode', base58_default_table, problem_table, content)


def base36_encode(content: str, problem_table: str = None):
    return switch_base_table_encode('b36encode', base36_default_table, problem_table, content)


def base32_encode(content: str, problem_table: str = None):
    return switch_base_table_encode('b32encode', base32_default_table, problem_table, content)


def base16_encode(content: str, problem_table: str = None):
    return switch_base_table_encode('b16encode', base16_default_table, problem_table, content)



import time
import socket
from typing import *
from ...logger import logger
from ..TcpClient import generate_free

from .EventArgs import NewConnectionEventArgs, MessageEventArgs,  CloseConnectionEventArgs
from .MessageHandler import MessageHanlder
from .BaseTcpHandler import BaseTcpHandler


class SimpleTcpService(BaseTcpHandler):

    def __handle_new_connection(self, connection: socket.socket, remote_host: str) -> None:
        logger.debug(
            f'TCPService:: connection establish from {remote_host}')
        if self.on_new_connection:
            accept = self.OnNewConnection(
                remote_host=remote_host, connection=connection)
            if not accept:
                return
        t = MessageHanlder(
            connection=connection,
            remote_host=remote_host,
            daemon=True,
            callback=self
        )
        self.threads.append(t)
        t.start()

    def run(self) -> None:
        super().run()
        logger.debug(
            f'TCPService:: server start listening on {self.host}:{self.port}')
        self.__on_listen(is_start=True)
        self.socket.listen(1)
        try:
            while self.is_listening and self.running:
                try:
                    connection, remote_host = self.socket.accept()
                except TimeoutError as e:
                    pass
                except Exception as e:
                    logger.info(f'TCPService:: listing stop for exception:{e}')
                self.__handle_new_connection(connection, remote_host)
        except:
            pass
        self.stop()
        self.__on_listen(is_start=False)

    def ensure_start(self):
        if self.is_listening:
            return
        if not self.running:
            self.start()
        counter = 100
        while not self.is_listening and counter:
            time.sleep(0.1)
            counter -= 1

    def __on_listen(self, is_start: bool):
        self.is_listening = is_start
        try:
            if is_start and self.on_listen_start:
                self.on_listen_start()
            elif not is_start and self.on_listen_end:
                self.on_listen_end()
        except:
            pass

    def stop(self):
        for t in self.threads:
            t.join()

    def __init__(self,
                 on_message: Callable[[MessageEventArgs], None],
                 on_new_connection: Callable[[
                     NewConnectionEventArgs], None] = None,
                 on_close_connection: Callable[[
                     CloseConnectionEventArgs], None] = None,
                 host: str = '127.0.0.1', port: int = None, ensure_start: bool = True):
        self.is_listening = False  # whether self-listening
        self.on_message: Callable = on_message
        self.on_new_connection: Callable = on_new_connection

        self.on_listen_start: Callable = None
        self.on_listen_end: Callable = None
        self.socket: socket.socket = None
        self.host = host
        self.running = False
        self.socket, self.port = generate_free(
            host=host,
            port=port,
        )
        super().__init__(
            on_message=on_message,
            on_new_connection=on_new_connection,
            on_close_connection=on_close_connection,
            daemon=True  # as daemon thread default
        )
        self.threads: List[MessageHanlder] = []
        if ensure_start:
            self.ensure_start()

    def start(self) -> None:
        if self.running:
            return
        self.running = True
        return super().start()



from .full_tree import BinaryTree
from typing import List, Type, overload
from .node import Node

class BinaryTree(BinaryTree):

    @staticmethod
    def from_full_tree(data) -> BinaryTree:
        t = BinaryTree(as_type=type(data))
        if data is None:
            return t
        data = BinaryTree.__convert_data(data)
        [t.add_as_full_tree(x) for x in data]
        return t

    @staticmethod
    def __from_full_tree_order(data: List, order: int = 0) -> BinaryTree:
        node = Node(None)
        pos = len(data) >> 1
        if pos == 0:
            node.data = data.pop(0)
            return node
        if order == 0:
            node.data = data.pop(0)
        node.lchild = BinaryTree.__from_full_tree_order(data[0:pos], order)
        if order == 1:
            node.data = data.pop(0)
        node.rchild = BinaryTree.__from_full_tree_order(data[pos:], order)
        if order == 2:
            node.data = data.pop(0)
        return node

    @staticmethod
    def from_full_tree_order(data: List, order: int = 0):
        as_type = type(data)
        data = BinaryTree.__convert_data(data)
        assert int(bin(len(data)+1)
                   [3:]) == 0, 'only data length to 2^n can be resolved'
        tree_data = BinaryTree.__from_full_tree_order(data, order)
        return BinaryTree(tree_data, as_type=as_type)

    @staticmethod
    def from_full_tree_preorder(data: List):
        return BinaryTree.from_full_tree_order(data, 0)

    @staticmethod
    def from_full_tree_inorder(data: List):
        return BinaryTree.from_full_tree_order(data, 1)

    @staticmethod
    def from_full_tree_postorder(data: List):
        return BinaryTree.from_full_tree_order(data, 2)



from typing import Callable, List


def list2sheet(lines: List, line_renderer: Callable = None, show_line_number: bool = True, max_show_count: int = 12, min_hide_count: int = 5, description: str = '     ...other {count} lines...'):
    '''
    show list in text mode
    '''
    result = [str(x) for x in lines if x]  # filter none
    result = [x for x in lines if len(x) > 0]  # filter empty
    count = len(result)
    skip_convert = None
    if count > max_show_count + min_hide_count:
        skip_convert = count - max_show_count
        half = int(max_show_count/2)
        result = result[0:half] + [None] + result[-half:]
    else:
        result = result

    def renderer(x: int):
        v = result[x]
        if not v:
            return str
        if line_renderer:
            v = line_renderer(v)
        if show_line_number:
            if skip_convert:
                line = f'{x if x<half else x-1:4d}: '
            else:
                line = f'{x:4d}: '
        else:
            line = ''
        content = f"{v if v else None}"
        return f'{line}{content}'

    result = [renderer(x) for x in range(len(result))]
    if skip_convert:
        result[half] = description.replace('{count}', str(skip_convert))
    return result



from Crypto.Cipher import AES, DES, DES3
from .utils import padding_to
from . import demo_3des, demo_aes, demo_blowfish, demo_des, demo_rsa


def main():
    key = bytes.fromhex('e738989a5bcb3c93a1101010')
    content = bytes.fromhex(
        '0c55dc63af806636fbc3bc81051c279c92ec9a6607aedaa877018cf4dbcc53cb2d')
    print(AES.new(padding_to(key, 32, 32), AES.MODE_ECB).decrypt(
        padding_to(content, 16)))
    config = {
        "key": "SGTSerfend2022",
        "iv": "1Ssecret1Ssecret1Ssecret"
    }
    content = ('0123456789abcdefAAAAAAAAAA'*5).encode()
    iv = str(config['iv']).encode()
    key = str(config['key']).encode()
    demo_des.main(key, iv, content)
    print('-'*20)
    demo_3des.main(key, iv, content)
    print('-'*20)
    demo_aes.main(key, iv, content)
    print('-'*20)
    demo_rsa.main(content)
    print('-'*20)
    demo_blowfish.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import os
import shutil
import sys

from setuptools import setup, find_packages
pck_name = 'sgtpyutils'
pck_dict = {}
pck_dict[pck_name] = pck_name
package_dir = os.path.dirname(os.path.realpath(__file__))
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()


def remove_dist(target: str):
    if os.path.exists(target):
        shutil.rmtree(target)


def load_about():
    about = {}
    pck_dir = os.path.join(package_dir, pck_name)
    with open(os.path.join(pck_dir, '__version__.py'), 'r', encoding='utf-8') as f:
        exec(f.read(), about)
    return about


def clear_dist():
    remove_dist('./build')
    remove_dist(f'./{pck_name}.egg-info')
    remove_dist(about['__public_path__'])


def load_requirements():
    if os.path.exists('requirements.txt'):
        install_requires = open('requirements.txt').read().splitlines()
    else:
        install_requires = []
    return install_requires


entry_points = {
    # 'console_scripts': [
    #     f'{pck_name} = {pck_name}:run',
    # ],
}

about = load_about()
install_requires = load_requirements()
print('install_requires', install_requires)
clear_dist()
setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    package_data={'': ['LICENSE', 'NOTICE']},
    package_dir=pck_dict,
    include_package_data=True,
    python_requires='>=3.7, <4',
    license=about['__license__'],
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries',
    ],
    keywords=about['__keywords__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(
        exclude=['tests', '*.tests', '*.tests.*', 'tests.*']),
    platforms='any',
    install_requires=install_requires,
    entry_points=entry_points
)


def upload():
    pub = about['__public_path__']
    arr = ['twine', 'upload', f'{pub}/*', '--verbose']
    p = os.system(' '.join(arr))


# if any([x.find('dist') > -1 for x in sys.argv]):
#     upload()



import random
from typing import Tuple
from ..logger import logger
import socket


def start_client(port: int, host: str = '127.0.0.1') -> socket.socket:
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    logger.info(f'start connect to local:{port}')
    client.connect((host, port))
    return client


def generate_free(host: str, port: int = None, timeout: int = 15, min_port: int = 1001, max_port: int = 60000) -> Tuple:
    '''
    get a free port if port not specify
    if port is occupied and port is specified , a exception would be raise
    if port is occupied and port is NOT specified , new try for random port would be used
    return Tuple[socket.socket, int]
    '''
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server.settimeout(timeout)
    r = random.Random()
    if not port:
        s_port = r.randint(min_port, max_port)
    else:
        s_port = port
    max_try_time = 1e3
    while max_try_time > 0:
        try:
            server.bind((host, s_port))
            break
        except Exception as e:
            if port:
                raise e
            s_port += 1
            if s_port >= int(max_port):
                s_port = min_port
            max_try_time -= 1

    return (server, s_port)



if str != bytes:
    # Python 3.x
    def ord(c):
        return c

    def chr(n):
        return bytes((n,))

table = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
table_size = len(table)
decode_table = dict((v, k) for k, v in enumerate(table))


def b58encode(v):
    int_val = 0
    for (i, c) in enumerate(v[::-1]):
        int_val += (256**i) * ord(c)

    result = []
    while int_val >= table_size:
        div, mod = divmod(int_val, table_size)
        result.append(table[mod])
        int_val = div
    result.append(table[int_val])

    padding = 0
    for c in v:
        if c == '\x00':
            padding += 1
        else:
            break

    result += [table[0]*padding]
    return ''.join(result)[::-1]


def b58decode(v):
    int_val = 0
    for (i, c) in enumerate(v[::-1]):
        int_val += decode_table[c] * (table_size**i)

    result = bytearray()
    while int_val >= 256:
        div, mod = divmod(int_val, 256)
        result.append(mod)
        int_val = div
    result.append(int_val)

    padding = 0
    for c in v:
        if c == table[0]:
            padding += 1
        else:
            break

    result += b'\x00' * padding
    return bytes(result)[::-1]



from typing import List


def cast2bytes(input: any, encoding: str = 'utf-8') -> bytes:
    if isinstance(input, List):
        return b''.join([cast2bytes(x, encoding) for x in input])
    if isinstance(input, str):
        return input.encode(encoding)
    if isinstance(input, int):
        data = hex(input)[2:]
        if len(data) % 2:
            data = f'0{data}'
        return bytes.fromhex(data)
    return bytes(input)



from .base_tree import BinaryTree
from typing import List, Type, overload
from .node import Node

class BinaryTree(BinaryTree):

    def add_as_full_tree(self, item):
        '''
        按满二叉树添加
        '''
        node = Node(item)
        if self.root is None:
            self.root = node
            return
        queue = [self.root]
        while queue:
            cur = queue.pop(0)
            if cur.lchild is None:
                cur.lchild = node
                return
            queue.append(cur.lchild)
            if cur.rchild is None:
                cur.rchild = node
                return
            queue.append(cur.rchild)

    def to_list(self):
        '''
        按从左到右，从上到下的顺序输出
        '''
        queue = [self.root]
        result = []
        while queue:
            cur = queue.pop(0)
            result.append(cur.data)
            if cur.lchild:
                queue.append(cur.lchild)
            if cur.rchild:
                queue.append(cur.rchild)
        return self.__convert_list(result)

    def __repr__(self) -> str:
        data = [str(x) for x in self.to_list()]
        return ''.join(data)




import struct

table = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789!#$%&()*+,./:;<=>?@[]^_`{|}~"'
table_size = len(table)
decode_table = dict((v, k) for k, v in enumerate(table))


def b91decode(data: str):
    if isinstance(data, bytes):
        print('excepted content to be a str , assume ascii encode.')
        data = data.decode('ascii')
    v = -1
    b = 0
    n = 0
    result = bytearray()
    for c in data:
        if not c in decode_table:
            continue
        c = decode_table[c]
        if(v < 0):
            v = c
        else:
            v += c * table_size
            b |= v << n
            n += 0xd if (v & 0x1fff) > 0x58 else 0xe
            while True:
                result.append(b & 0xff)
                b >>= 8
                n -= 8
                if not n > 7:
                    break
            v = -1
    if v + 1:
        result.append((b | v << n) & 0xff)
    return bytes(result)


def b91encode(data: bytes):
    if isinstance(data, str):
        print('excepted content to be a bytes , assume utf-8 encode.')
        data = data.encode('utf-8')
    data = bytearray(data)
    b = 0
    n = 0
    result = []
    for count in range(len(data)):
        b |= data[count] << n
        n += 8
        if n > 13:
            v = b & 0x1fff
            if v > 0x58:
                b >>= 13
                n -= 13
            else:
                v = b & 0x3fff
                b >>= 14
                n -= 14
            result.append(table[v % table_size] + table[v // table_size])
    if n:
        result.append(table[b % table_size])
        if n > 7 or b >= table_size:
            result.append(table[b // table_size])
    return ''.join(result)



from itertools import *
from typing import List


def get_end_by_length(table: any, length: int):
    return len(table) ** length


def run_cycle(table: List, end: int = -1, length: int = -1, start: int = 0):
    '''
    get a iter of tables-cycle
    Arguments:
    table:str|List the table of string-collectin
    end:int end of iter, if not set , one-cycle of table will return
    length:int length of cycle-string
      if length is set , `end` will be ignored
    start:int from what to return
    Example:
      > import string
      > a = run_cycle(string.digits,10)
      > for index,i in a:
      >   print(index,''.join(i))

      > 0 0
        1 1
        2 2
        3 3
        4 4
        5 5
        6 6
        7 7
        8 8
        9 9

      > a = run_cycle('123abc456',10)
      > for index,i in a:
      >   print(i)

      > (0, ['1', '1'])
        (1, ['2', '1'])
        (2, ['3', '1'])
        (3, ['a', '1'])
        (4, ['b', '1'])
        (5, ['c', '1'])
        (6, ['4', '1'])
        (7, ['5', '1'])
        (8, ['6', '1'])
        (9, ['1', '2'])

      > import string
      > a = run_cycle(string.digits,length=1)
      > for index,i in a:
      >   print(index,''.join(i))

      > 0 0
        1 1
        2 2
        3 3
        4 4
        5 5
        6 6
        7 7
        8 8
        9 9
    '''
    if length > 0:
        end = get_end_by_length(table, length)
    if isinstance(table, str):
        table = [x for x in table]
    table_size = len(table)

    def int2table(v: int) -> List:
        r = []
        while v >= table_size:
            r.append(v % table_size)
            v //= table_size
        r.append(v)
        return r

    table_start = int2table(start)
    table_end = int2table(end-1)
    table_start += [0]*(len(table_end)-len(table_start))

    base_str = [table[x] for x in table_start]

    def accumulate(index: int = 0):
        table_start[index] += 1
        if table_start[index] >= table_size:
            table_start[index] = 0
            base_str[index] = table[0]
            return accumulate(index+1)
        base_str[index] = table[table_start[index]]

    yield 0, base_str
    for i in range(start+1, end):
        accumulate()
        yield i, base_str



from Crypto.Cipher import AES, DES, DES3
from .utils import padding_to


def main(key: bytes, iv: bytes, content: bytes):
    # MODE_CBC
    e = DES3.new(padding_to(key, 24, 24), DES3.MODE_CBC,
                 padding_to(iv, 8, 8)).encrypt(padding_to(content, 8))
    d = DES3.new(padding_to(key, 24, 24), DES3.MODE_CBC,
                 padding_to(iv, 8, 8)).decrypt(e)
    print(f'e = {e}\nd = {d}\n\n')
    # MODE_CFB
    e = DES3.new(padding_to(key, 24, 24), DES3.MODE_CFB,
                 padding_to(iv, 8, 8)).encrypt(padding_to(content, 8))
    d = DES3.new(padding_to(key, 24, 24), DES3.MODE_CFB,
                 padding_to(iv, 8, 8)).decrypt(e)
    print(f'e = {e}\nd = {d}\n\n')
    # MODE_ECB no need for iv
    e = DES3.new(padding_to(key, 24, 24), DES3.MODE_ECB).encrypt(
        padding_to(content, 8))
    d = DES3.new(padding_to(key, 24, 24), DES3.MODE_ECB).decrypt(e)
    print(f'e = {e}\nd = {d}\n\n')
    # MODE_OFB
    e = DES3.new(padding_to(key, 24, 24), DES3.MODE_OFB,
                 padding_to(iv, 8, 8)).encrypt(padding_to(content, 8))
    d = DES3.new(padding_to(key, 24, 24), DES3.MODE_OFB,
                 padding_to(iv, 8, 8)).decrypt(e)
    print(f'e = {e}\nd = {d}\n\n')
